@@ -1,21 +1,26 @@
 ###############################################################################
-# 
+#
 # configglue -- glue for your apps' configuration
-# 
+#
 # A library for simple, DRY configuration of applications
-# 
+#
 # (C) 2009--2010 by Canonical Ltd.
 # originally by John R. Lenton <john.lenton@canonical.com>
 # incorporating schemaconfig as configglue.pyschema
 # schemaconfig originally by Ricardo Kirkner <ricardo.kirkner@canonical.com>
-# 
+#
 # Released under the BSD License (see the file LICENSE)
-# 
+#
 # For bug reports, support, and new releases: http://launchpad.net/configglue
-# 
+#
 ###############################################################################
 
+import os
 import sys
+from ConfigParser import (
+    NoOptionError,
+    NoSectionError,
+)
 from optparse import OptionParser
 from collections import namedtuple
 
@@ -34,7 +39,7 @@ SchemaGlue = namedtuple("SchemaGlue",
 
 def schemaconfigglue(parser, op=None, argv=None):
     """Glue an OptionParser with a SchemaConfigParser.
-    
+
     The OptionParser is populated with options and defaults taken from the
     SchemaConfigParser.
 
@@ -62,21 +67,36 @@ def schemaconfigglue(parser, op=None, argv=None):
             kwargs = {}
             if option.help:
                 kwargs['help'] = option.help
-            kwargs['default'] = parser.get(section.name, option.name)
+            if not option.fatal:
+                kwargs['default'] = parser.get(section.name, option.name)
             kwargs['action'] = option.action
             og.add_option('--' + long_name(option), **kwargs)
     options, args = op.parse_args(argv)
 
+    def set_value(section, option, value):
+        # if value is not of the right type, cast it
+        if not option.validate(value):
+            value = option.parse(value)
+        parser.set(section.name, option.name, value)
+
     for section in schema.sections():
         for option in section.options():
-            value = getattr(options, opt_name(option))
-            if parser.get(section.name, option.name) != value:
-                # the value has been overridden by an argument;
-                # update it, but make sure it's a string, as
-                # SafeConfigParser will complain otherwise.
-                if not isinstance(value, basestring):
-                    value = repr(value)
-                parser.set(section.name, option.name, value)
+            # 1. op value != parser value
+            # 2. op value == parser value != env value
+            # 3. op value == parser value == env value or not env value
+
+            op_value = getattr(options, opt_name(option))
+            try:
+                parser_value = parser.get(section.name, option.name)
+            except (NoSectionError, NoOptionError):
+                parser_value = None
+            env_value = os.environ.get("CONFIGGLUE_{0}".format(
+                long_name(option).upper()))
+
+            if op_value != parser_value:
+                set_value(section, option, op_value)
+            elif env_value is not None and env_value != parser_value:
+                set_value(section, option, env_value)
 
     return op, options, args
 
@@ -99,6 +119,5 @@ def configglue(schema_class, configs, usage=None):
     parser, opts, args = schemaconfigglue(scp, op=op)
     is_valid, reasons = scp.is_valid(report=True)
     if not is_valid:
-        parser.error(reasons[0])
+        parser.error('\n'.join(reasons))
     return SchemaGlue(scp, parser, opts, args)
-
