@@ -1,24 +1,25 @@
 ###############################################################################
-# 
+#
 # configglue -- glue for your apps' configuration
-# 
+#
 # A library for simple, DRY configuration of applications
-# 
+#
 # (C) 2009--2010 by Canonical Ltd.
 # originally by John R. Lenton <john.lenton@canonical.com>
 # incorporating schemaconfig as configglue.pyschema
 # schemaconfig originally by Ricardo Kirkner <ricardo.kirkner@canonical.com>
-# 
+#
 # Released under the BSD License (see the file LICENSE)
-# 
+#
 # For bug reports, support, and new releases: http://launchpad.net/configglue
-# 
+#
 ###############################################################################
 
 import codecs
 import collections
 import copy
 import os
+import re
 import string
 
 from ConfigParser import (
@@ -29,7 +30,7 @@ from ConfigParser import (
     NoSectionError,
 )
 
-from configglue.pyschema.schema import DictConfigOption
+from configglue.pyschema.schema import DictOption
 
 
 __all__ = [
@@ -88,13 +89,19 @@ class SchemaConfigParser(BaseConfigParser, object):
             schema_sections.update(skip_sections)
             sections_match = (parsed_sections == schema_sections)
             if not sections_match:
-                error_msg = "Sections in configuration do not match schema: %s"
                 unmatched_sections = list(parsed_sections - schema_sections)
-                error_value = ', '.join(unmatched_sections)
-                errors.append(error_msg % error_value)
+                if len(unmatched_sections) > 0:
+                    error_msg = "Sections in configuration are missing from schema: %s"
+                    error_value = ', '.join(unmatched_sections)
+                    errors.append(error_msg % error_value)
+                unmatched_sections = list(schema_sections - parsed_sections)
+                if len(unmatched_sections) > 0:
+                    error_msg = "Sections in schema are missing from configuration: %s"
+                    error_value = ', '.join(unmatched_sections)
+                    errors.append(error_msg % error_value)
             valid &= sections_match
 
-            for name in parsed_sections:
+            for name in parsed_sections.union(schema_sections):
                 if name not in skip_sections:
                     if not self.schema.has_section(name):
                         # this should have been reported before
@@ -102,7 +109,10 @@ class SchemaConfigParser(BaseConfigParser, object):
                         continue
 
                     section = self.schema.section(name)
-                    parsed_options = set(self.options(name))
+                    try:
+                        parsed_options = set(self.options(name))
+                    except NoSectionError:
+                        parsed_options = set([])
                     schema_options = set(section.options())
 
                     fatal_options = set(opt.name for opt in schema_options
@@ -138,9 +148,8 @@ class SchemaConfigParser(BaseConfigParser, object):
             self.parse_all()
 
         except Exception, e:
-            if valid:
-                errors.append(e)
-                valid = False
+            errors.append(str(e))
+            valid = False
 
         if report:
             return valid, errors
@@ -204,7 +213,7 @@ class SchemaConfigParser(BaseConfigParser, object):
         If section is specified, return all options from that section only.
 
         Section is to be specified *by name*, not by
-        passing in real ConfigSection objects.
+        passing in real Section objects.
 
         """
         values = collections.defaultdict(dict)
@@ -327,9 +336,9 @@ class SchemaConfigParser(BaseConfigParser, object):
                 kwargs = {'parser': self}
 
             # hook to save extra sections
-            is_dict_option = isinstance(option_obj, DictConfigOption)
+            is_dict_option = isinstance(option_obj, DictOption)
             is_dict_lines_option = (hasattr(option_obj, 'item') and
-                isinstance(option_obj.item, DictConfigOption))
+                isinstance(option_obj.item, DictOption))
             is_default_value = unicode(option_obj.default) == value
 
             # avoid adding implicit sections for dict default value
@@ -349,7 +358,7 @@ class SchemaConfigParser(BaseConfigParser, object):
 
             if is_default_value:
                 value = option_obj.default
-            else:
+            elif isinstance(value, basestring):
                 try:
                     value = option_obj.parse(value, **kwargs)
                 except ValueError, e:
@@ -365,8 +374,8 @@ class SchemaConfigParser(BaseConfigParser, object):
         If any options are omitted from the config file, provide the
         default value from the schema.
 
-        In the case of an NoSectionError or NoOptionError, raise it if the option
-        has *fatal* set to *True*.
+        In the case of an NoSectionError or NoOptionError, raise it if the
+        option has *fatal* set to *True*.
 
         """
         for section in self.schema.sections():
@@ -381,17 +390,18 @@ class SchemaConfigParser(BaseConfigParser, object):
         """Return the location (file) where the option was last defined."""
         return self._location.get(option)
 
+    def _extract_interpolation_keys(self, item):
+        if isinstance(item, (list, tuple)):
+            keys = map(self._extract_interpolation_keys, item)
+            keys = reduce(set.union, keys, set())
+        else:
+            keys = set(self._KEYCRE.findall(item))
+        # remove invalid key
+        if '' in keys:
+            keys.remove('')
+        return keys
+
     def _get_interpolation_keys(self, section, option):
-        def extract_keys(item):
-            if isinstance(item, (list, tuple)):
-                keys = map(extract_keys, item)
-                keys = reduce(set.union, keys, set())
-            else:
-                keys = set(self._KEYCRE.findall(item))
-            # remove invalid key
-            if '' in keys:
-                keys.remove('')
-            return keys
 
         rawval = super(SchemaConfigParser, self).get(section, option, True)
         try:
@@ -400,7 +410,7 @@ class SchemaConfigParser(BaseConfigParser, object):
         except:
             value = rawval
 
-        keys = extract_keys(value)
+        keys = self._extract_interpolation_keys(value)
         return rawval, keys
 
     def _interpolate_value(self, section, option):
@@ -435,38 +445,36 @@ class SchemaConfigParser(BaseConfigParser, object):
         assert isinstance(result, basestring)
         return result
 
+    def interpolate_environment(self, rawval, raw=False):
+        if raw:
+            return rawval
+
+        # interpolate environment variables
+        pattern = re.sub(r'\${([A-Z_]+)}', r'%(\1)s', rawval)
+        pattern = re.sub(r'\$([A-Z_]+)', r'%(\1)s', pattern)
+
+        keys = self._extract_interpolation_keys(pattern)
+        if not keys:
+            # interpolation keys are not valid
+            return rawval
+
+        interpolated = pattern % os.environ
+        return interpolated
+
     def _get_default(self, section, option):
-        # mark the value as not initialized to be able to have a None default
-        marker = object()
-        value = marker
-
         # cater for 'special' sections
-        if section == '__main__':
-            opt = getattr(self.schema, option, None)
-            if opt is not None and not opt.fatal:
-                value = opt.default
-        elif section == '__noschema__':
+        if section == '__noschema__':
             value = super(SchemaConfigParser, self).get(section, option)
-        else:
-            try:
-                opt = self.schema.section(section).option(option)
-                if not opt.fatal:
-                    value = opt.default
-            except Exception:
-                pass
+            return value
 
-        if value is marker:
-            # value was not set, so either section or option was not found
-            # or option was required (fatal set to True)
-            #if self.schema.has_section(section):
-            #    raise NoOptionError(option, section)
-            #else:
-            #    raise NoSectionError(section)
-            return None
-        else:
-            # we want to return a non-parsed value
-            # a unicode of the value is the closest we can get
-            return unicode(value)
+        # any other section
+        opt = self.schema.section(section).option(option)
+        if not opt.fatal:
+            value = opt.default
+            return value
+
+        # no default value found, raise an error
+        raise NoOptionError(option, section)
 
     def get(self, section, option, raw=False, vars=None, parse=True):
         """Return the parsed value of an option.
@@ -482,8 +490,7 @@ class SchemaConfigParser(BaseConfigParser, object):
         try:
             # get option's raw mode setting
             try:
-                section_obj = self.schema.section(section)
-                option_obj = section_obj.option(option)
+                option_obj = self._get_option(section, option)
                 raw = option_obj.raw or raw
             except:
                 pass
@@ -500,26 +507,45 @@ class SchemaConfigParser(BaseConfigParser, object):
             # option not found in config, try to get its default value from
             # schema
             value = self._get_default(section, option)
-            if value is None:
-                raise
 
             # value found, so section and option exist
             # add it to the config
-            if not self.has_section(section):
-                # Don't call .add_section here because 2.6 complains
-                # about sections called '__main__'
-                self._sections[section] = {}
             self.set(section, option, value)
 
-        if parse:
-            value = self.parse(section, option, value)
+        # interpolate environment variables
+        if isinstance(value, basestring):
+            try:
+                value = self.interpolate_environment(value, raw=raw)
+                if parse:
+                    value = self.parse(section, option, value)
+            except KeyError:
+                # interpolation failed, fallback to default value
+                value = self._get_default(section, option)
+
         return value
 
+    def _get_option(self, section, option):
+        section_obj = self.schema.section(section)
+        option_obj = section_obj.option(option)
+        return option_obj
+
     def set(self, section, option, value):
         """Set an option's raw value."""
-        super(SchemaConfigParser, self).set(section, option, value)
+        option_obj = self._get_option(section, option)
+        # make sure the value is of the right type for the option
+        if not option_obj.validate(value):
+            raise TypeError("{0} is not a valid {1} value.".format(
+                value, type(option_obj).__name__))
+        # cast value to a string because SafeConfigParser only allows
+        # strings to be set
+        str_value = option_obj.to_string(value)
+        if not self.has_section(section):
+            # Don't call .add_section here because 2.6 complains
+            # about sections called '__main__'
+            self._sections[section] = {}
+        super(SchemaConfigParser, self).set(section, option, str_value)
         filename = self.locate(option)
-        self._dirty[filename][section][option] = value
+        self._dirty[filename][section][option] = str_value
 
     def write(self, fp):
         """Write an .ini-format representation of the configuration state."""
@@ -568,4 +594,3 @@ class SchemaConfigParser(BaseConfigParser, object):
                     os.rename(filename, "%s.old" % filename)
                 # rename new file
                 os.rename("%s.new" % filename, filename)
-
