@@ -1,68 +1,131 @@
 ###############################################################################
-# 
+#
 # configglue -- glue for your apps' configuration
-# 
+#
 # A library for simple, DRY configuration of applications
-# 
+#
 # (C) 2009--2010 by Canonical Ltd.
 # originally by John R. Lenton <john.lenton@canonical.com>
 # incorporating schemaconfig as configglue.pyschema
 # schemaconfig originally by Ricardo Kirkner <ricardo.kirkner@canonical.com>
-# 
+#
 # Released under the BSD License (see the file LICENSE)
-# 
+#
 # For bug reports, support, and new releases: http://launchpad.net/configglue
-# 
+#
 ###############################################################################
 
+from ConfigParser import (
+    NoSectionError,
+    NoOptionError,
+)
 from copy import deepcopy
 from inspect import getmembers
+from warnings import warn
 
 
 __all__ = [
     'BoolConfigOption',
+    'BoolOption',
     'ConfigOption',
+    'Option',
     'ConfigSection',
+    'Section',
     'DictConfigOption',
+    'DictOption',
     'IntConfigOption',
+    'IntOption',
     'LinesConfigOption',
+    'ListOption',
     'Schema',
     'StringConfigOption',
+    'StringOption',
     'TupleConfigOption',
+    'TupleOption',
+    'merge',
 ]
 
 NO_DEFAULT = object()
 
-_internal = object.__dict__.keys() + ['__module__']
-
 
 def get_config_objects(obj):
-    objects = ((n, o) for (n, o) in getmembers(obj)
-        if isinstance(o, (ConfigSection, ConfigOption)))
+    """Return the list of Section- and Option-derived objects."""
+    objects = []
+    for name, obj in getmembers(obj):
+        if isinstance(obj, (Section, Option)):
+            objects.append((name, obj))
+        elif type(obj) == type and issubclass(obj, Section):
+            instance = obj()
+            for key, value in get_config_objects(obj):
+                setattr(instance, key, value)
+            objects.append((name, instance))
     return objects
 
 
+def merge(*schemas):
+    # import here to avoid circular imports
+    from .parser import SchemaValidationError
+
+    # define result schema
+    class MergedSchema(Schema):
+        pass
+
+    # for each schema
+    for schema in schemas:
+        instance = schema()
+        # iterate over the schema sections
+        for section in instance.sections():
+            # create the appropriate section object
+            section_name = section.name
+            if section_name == '__main__':
+                # section is special __main__ section
+                sect = MergedSchema
+            elif not hasattr(MergedSchema, section_name):
+                # section is not present in schema, just copy it over
+                # completely
+                setattr(MergedSchema, section_name, section)
+                continue
+            else:
+                # section is present in schema, do the merge
+                sect = getattr(MergedSchema, section_name)
+
+            # do the merge
+
+            # iterate over the section options
+            for option in section.options():
+                opt = getattr(sect, option.name, None)
+                if opt is not None:
+                    if option != opt:
+                        raise SchemaValidationError("Conflicting option "
+                            "'%s.%s' while merging schemas." % (
+                            section_name, option.name))
+                else:
+                    setattr(sect, option.name, option)
+
+    return MergedSchema
+
+
 class Schema(object):
     """A complete description of a system configuration.
 
     To define your own configuration schema you should:
      1- Inherit from Schema
-     2- Add ConfigOptions and ConfigSections as class attributes.
+     2- Add Options and Sections as class attributes.
 
     With that your whole configuration schema is defined, and you can now
     load configuration files.
 
-    ConfigOptions that don't go in a ConfigSection will belong in the
+    Options that don't go in a Section will belong in the
     '__main__' section of the configuration files.
 
-    One ConfigOption comes already defined in Schema, 'includes' in the
+    One Option comes already defined in Schema, 'includes' in the
     '__main__' section, that allows configuration files to include other
     configuration files.
 
     """
 
     def __init__(self):
-        self.includes = LinesConfigOption(item=StringConfigOption())
+        self.includes = ListOption(item=StringOption())
         self._sections = {}
         # add section and options to the schema
         for name, item in get_config_objects(self.__class__):
@@ -71,9 +134,9 @@ class Schema(object):
     def _add_item(self, name, item):
         """Add a top-level item to the schema."""
         item.name = name
-        if isinstance(item, ConfigSection):
+        if isinstance(item, Section):
             self._add_section(name, item)
-        elif isinstance(item, ConfigOption):
+        elif isinstance(item, Option):
             self._add_option(name, item)
         # override class attributes with instance attributes to correctly
         # handle schema inheritance
@@ -88,38 +151,44 @@ class Schema(object):
 
     def _add_option(self, name, option):
         """Add a top-level option to the schema."""
-        section = self._sections.setdefault('__main__', ConfigSection(name='__main__'))
+        section = self._sections.setdefault('__main__',
+            Section(name='__main__'))
         option.section = section
         setattr(section, name, option)
 
     def __eq__(self, other):
-        return (self._sections == other._sections and
-                self.includes == other.includes)
+        return (
+            type(self) == type(other) and
+            self._sections == other._sections and
+            self.includes == other.includes)
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
 
     def is_valid(self):
         """Return whether the schema has a valid structure."""
         explicit_default_section = isinstance(getattr(self, '__main__', None),
-                                              ConfigSection)
+                                              Section)
         is_valid = not explicit_default_section
         return is_valid
 
     def has_section(self, name):
-        """Return whether the schema as a given section."""
-        """Return True if a ConfigSection with the given name is available"""
+        """Return True if a Section with the given name is available"""
         return name in self._sections.keys()
 
     def section(self, name):
-        """Return a ConfigSection by name"""
+        """Return a Section by name"""
         section = self._sections.get(name)
-        assert section is not None, "Invalid ConfigSection name '%s'" % name
+        if section is None:
+            raise NoSectionError(name)
         return section
 
     def sections(self):
-        """Returns the list of available ConfigSections"""
+        """Returns the list of available Sections"""
         return self._sections.values()
 
     def options(self, section=None):
-        """Return all the ConfigOptions within a given section.
+        """Return all the Options within a given section.
 
         If section is omitted, returns all the options in the configuration
         file, flattening out any sections.
@@ -135,18 +204,18 @@ class Schema(object):
         elif section.name == '__main__':
             class_config_objects = get_config_objects(self.__class__)
             options = [getattr(self, att) for att, _ in class_config_objects
-                           if isinstance(getattr(self, att), ConfigOption)]
+                           if isinstance(getattr(self, att), Option)]
         else:
             options = section.options()
         return options
 
 
-class ConfigSection(object):
+class Section(object):
     """A group of options.
 
-    This class is just a bag you can dump ConfigOptions in.
+    This class is just a bag you can dump Options in.
 
-    After instantiating the Schema, each ConfigSection will know its own
+    After instantiating the Schema, each Section will know its own
     name.
 
     """
@@ -154,38 +223,44 @@ class ConfigSection(object):
         self.name = name
 
     def __eq__(self, other):
-        return (self.name == other.name and
-                self.options() == other.options())
+        return (
+            type(self) == type(other) and
+            self.name == other.name and
+            self.options() == other.options())
+
+    def __ne__(self, other):
+        return not self.__eq__(other)
 
     def __repr__(self):
         if self.name:
             name = " %s" % self.name
         else:
             name = ''
-        value = "<ConfigSection%s>" % name
+        value = "<{0}{1}>".format(self.__class__.__name__, name)
         return value
 
     def has_option(self, name):
-        """Return True if a ConfigOption with the given name is available"""
+        """Return True if a Option with the given name is available"""
         opt = getattr(self, name, None)
-        return isinstance(opt, ConfigOption)
+        return isinstance(opt, Option)
 
     def option(self, name):
-        """Return a ConfigOption by name"""
+        """Return a Option by name"""
         opt = getattr(self, name, None)
-        assert opt is not None, "Invalid ConfigOption name '%s'" % name
+        if opt is None:
+            raise NoOptionError(name, self.name)
         return opt
 
     def options(self):
-        """Return a list of all available ConfigOptions within this section"""
+        """Return a list of all available Options within this section"""
         return [getattr(self, att) for att in vars(self)
-                if isinstance(getattr(self, att), ConfigOption)]
+                if isinstance(getattr(self, att), Option)]
 
 
-class ConfigOption(object):
+class Option(object):
     """Base class for Config Options.
 
-    ConfigOptions are never bound to a particular conguration file, and
+    Options are never bound to a particular conguration file, and
     simply describe one particular available option.
 
     They also know how to parse() the content of a config file in to the right
@@ -198,22 +273,22 @@ class ConfigOption(object):
     argument, parser, that should receive the whole SchemaConfigParser to
     do the parsing.  This is needed for config options that need to look at
     other parts of the config file to be able to carry out their parsing,
-    like DictConfigOptions.
+    like DictOption.
 
     If self.fatal == True, SchemaConfigParser's parse_all will raise an
     exception if no value for this option is provided in the configuration
     file.  Otherwise, the self.default value will be used if the option is
     omitted.
 
-    In runtime, after instantiating the Schema, each ConfigOption will also
+    In runtime, after instantiating the Schema, each Option will also
     know its own name and to which section it belongs.
 
     """
 
     require_parser = False
 
-    def __init__(self, name='', raw=False, default=NO_DEFAULT, fatal=False, help='',
-                 section=None, action='store'):
+    def __init__(self, name='', raw=False, default=NO_DEFAULT, fatal=False,
+                 help='', section=None, action='store'):
         self.name = name
         self.raw = raw
         self.fatal = fatal
@@ -226,11 +301,13 @@ class ConfigOption(object):
 
     def __eq__(self, other):
         try:
-            equal = (self.name == other.name and
-                     self.raw == other.raw and
-                     self.fatal == other.fatal and
-                     self.default == other.default and
-                     self.help == other.help)
+            equal = (
+                type(self) == type(other) and
+                self.name == other.name and
+                self.raw == other.raw and
+                self.fatal == other.fatal and
+                self.default == other.default and
+                self.help == other.help)
             if self.section is not None and other.section is not None:
                 # only test for section name to avoid recursion
                 equal &= self.section.name == other.section.name
@@ -241,17 +318,20 @@ class ConfigOption(object):
 
         return equal
 
+    def __ne__(self, other):
+        return not self.__eq__(other)
+
     def __repr__(self):
         extra = ' raw' if self.raw else ''
         extra += ' fatal' if self.fatal else ''
         section = self.section.name if self.section is not None else None
         if section is not None:
-            name = " %s.%s" % (section, self.name)
+            name = " {0}.{1}".format(section, self.name)
         elif self.name:
-            name = " %s" % self.name
+            name = " {0}".format(self.name)
         else:
             name = ''
-        value = "<ConfigOption%s%s>" % (name, extra)
+        value = "<{0}{1}{2}>".format(self.__class__.__name__, name, extra)
         return value
 
     def _get_default(self):
@@ -261,9 +341,16 @@ class ConfigOption(object):
         """Parse the given value."""
         raise NotImplementedError()
 
+    def validate(self, value):
+        raise NotImplementedError()
 
-class BoolConfigOption(ConfigOption):
-    """A ConfigOption that is parsed into a bool"""
+    def to_string(self, value):
+        """Return a string representation of the value."""
+        return str(value)
+
+
+class BoolOption(Option):
+    """A Option that is parsed into a bool"""
 
     def _get_default(self):
         return False
@@ -284,9 +371,12 @@ class BoolConfigOption(ConfigOption):
         else:
             raise ValueError("Unable to determine boolosity of %r" % value)
 
+    def validate(self, value):
+        return isinstance(value, bool)
 
-class IntConfigOption(ConfigOption):
-    """A ConfigOption that is parsed into an int"""
+
+class IntOption(Option):
+    """A Option that is parsed into an int"""
 
     def _get_default(self):
         return 0
@@ -302,13 +392,16 @@ class IntConfigOption(ConfigOption):
 
         return int(value)
 
+    def validate(self, value):
+        return isinstance(value, int)
 
-class LinesConfigOption(ConfigOption):
-    """A ConfigOption that is parsed into a list of objects
+
+class ListOption(Option):
+    """A Option that is parsed into a list of objects
 
     All items in the list need to be of the same type.  The 'item' constructor
     argument determines the type of the list items. item should be another
-    child of ConfigOption.
+    child of Option.
 
     self.require_parser will be True if the item provided in turn has
     require_parser == True.
@@ -321,7 +414,7 @@ class LinesConfigOption(ConfigOption):
 
     def __init__(self, name='', item=None, raw=False, default=NO_DEFAULT,
         fatal=False, help='', action='store', remove_duplicates=False):
-        super(LinesConfigOption, self).__init__(name=name, raw=raw,
+        super(ListOption, self).__init__(name=name, raw=raw,
             default=default, fatal=fatal, help=help, action=action)
         self.item = item
         self.require_parser = item.require_parser
@@ -353,9 +446,12 @@ class LinesConfigOption(ConfigOption):
             items = filtered_items
         return items
 
+    def validate(self, value):
+        return isinstance(value, list)
 
-class StringConfigOption(ConfigOption):
-    """A ConfigOption that is parsed into a string.
+
+class StringOption(Option):
+    """A Option that is parsed into a string.
 
     If null==True, a value of 'None' will be parsed in to None instead of
     just leaving it as the string 'None'.
@@ -365,7 +461,7 @@ class StringConfigOption(ConfigOption):
     def __init__(self, name='', raw=False, default=NO_DEFAULT, fatal=False,
         null=False, help='', action='store'):
         self.null = null
-        super(StringConfigOption, self).__init__(name=name, raw=raw,
+        super(StringOption, self).__init__(name=name, raw=raw,
             default=default, fatal=fatal, help=help, action=action)
 
     def _get_default(self):
@@ -387,9 +483,17 @@ class StringConfigOption(ConfigOption):
             result = repr(value)
         return result
 
+    def to_string(self, value):
+        if value is None and self.null:
+            return 'None'
+        return value
 
-class TupleConfigOption(ConfigOption):
-    """A ConfigOption that is parsed into a fixed-size tuple of strings.
+    def validate(self, value):
+        return (self.null and value is None) or isinstance(value, basestring)
+
+
+class TupleOption(Option):
+    """A Option that is parsed into a fixed-size tuple of strings.
 
     The number of items in the tuple should be specified with the 'length'
     constructor argument.
@@ -398,7 +502,7 @@ class TupleConfigOption(ConfigOption):
 
     def __init__(self, name='', length=0, raw=False, default=NO_DEFAULT,
         fatal=False, help='', action='store'):
-        super(TupleConfigOption, self).__init__(name=name, raw=raw,
+        super(TupleOption, self).__init__(name=name, raw=raw,
             default=default, fatal=fatal, help=help, action=action)
         self.length = length
 
@@ -419,15 +523,19 @@ class TupleConfigOption(ConfigOption):
             if len(parts) == self.length:
                 result = tuple(parts)
             else:
-                raise ValueError("Tuples need to be %d items long" % self.length)
+                raise ValueError(
+                    "Tuples need to be %d items long" % self.length)
         else:
             result = tuple(parts)
             # length is 0, so no length validation
         return result
 
+    def validate(self, value):
+        return isinstance(value, tuple)
 
-class DictConfigOption(ConfigOption):
-    """A ConfigOption that is parsed into a dictionary.
+
+class DictOption(Option):
+    """A Option that is parsed into a dictionary.
 
     In the configuration file you'll need to specify the name of a section,
     and all that section's items will be parsed as a dictionary.
@@ -435,7 +543,7 @@ class DictConfigOption(ConfigOption):
     The available keys for the dict are specified with the 'spec' constructor
     argument, that should be in turn a dictionary.  spec's keys are the
     available keys for the config file, and spec's values should be
-    ConfigOptions that will be used to parse the values in the config file.
+    Options that will be used to parse the values in the config file.
 
     """
 
@@ -447,11 +555,11 @@ class DictConfigOption(ConfigOption):
         if spec is None:
             spec = {}
         if item is None:
-            item = StringConfigOption()
+            item = StringOption()
         self.spec = spec
         self.strict = strict
         self.item = item
-        super(DictConfigOption, self).__init__(name=name, raw=raw,
+        super(DictOption, self).__init__(name=name, raw=raw,
             default=default, fatal=fatal, help=help, action=action)
 
     def _get_default(self):
@@ -504,20 +612,31 @@ class DictConfigOption(ConfigOption):
                     result[key] = value
         return result
 
+    def validate(self, value):
+        return isinstance(value, dict)
+
     def get_extra_sections(self, section, parser):
+        """Return the list of implicit sections.
+
+        Implicit sections are sections defined in the configuration file
+        that are not defined in the schema, but used as helper sections for
+        defining a dictionary.
+
+        """
         sections = []
         for option in parser.options(section):
             option_obj = self.spec.get(option, self.item)
-            is_dict_item = isinstance(option_obj, DictConfigOption)
+            is_dict_item = isinstance(option_obj, DictOption)
             is_dict_lines_item = (hasattr(option_obj, 'item') and
-                isinstance(option_obj.item, DictConfigOption))
+                isinstance(option_obj.item, DictOption))
+
+            if not (is_dict_item or is_dict_lines_item):
+                continue
 
             if is_dict_item:
                 base = option_obj
-            elif is_dict_lines_item:
-                base = option_obj.item
             else:
-                continue
+                base = option_obj.item
 
             value = parser.get(section, option, parse=False)
             names = value.split()
@@ -530,3 +649,45 @@ class DictConfigOption(ConfigOption):
 
         return sections
 
+#
+# deprecated
+#
+
+
+class Deprecated(type):
+    def __init__(cls, name, bases, attrs):
+        warn('{0} is deprecated; use {1} instead.'.format(
+            name, bases[0].__name__), DeprecationWarning)
+        type.__init__(cls, name, bases, attrs)
+
+
+class StringConfigOption(StringOption):
+    __metaclass__ = Deprecated
+
+
+class IntConfigOption(IntOption):
+    __metaclass__ = Deprecated
+
+
+class BoolConfigOption(BoolOption):
+    __metaclass__ = Deprecated
+
+
+class DictConfigOption(DictOption):
+    __metaclass__ = Deprecated
+
+
+class LinesConfigOption(ListOption):
+    __metaclass__ = Deprecated
+
+
+class TupleConfigOption(TupleOption):
+    __metaclass__ = Deprecated
+
+
+class ConfigOption(Option):
+    __metaclass__ = Deprecated
+
+
+class ConfigSection(Section):
+    __metaclass__ = Deprecated
